import waba.fx.*;   // Graphics
//import waba.io.*;
import waba.sys.*;  // Convert
import waba.ui.*;
//import waba.util.*;
import extra.ui.*;  // wextras: List (Popup, RelativeContainer, PreferredSize), Title
// Random

/**
 * @author <A HREF="mailto:weyer@kagi.com">Steve Weyer</A>
 * Copyright (c) 2001, S. Weyer. All right reserved.
 * @version waba 1.0
 * !!!b 08 Feb 2001: moved appG init; patternPicker init w/o Vector; setXY fill area; a few more patterns
 * !!!a 17 Jan 2001
 * <A HREF="http://www.kagi.com/weyer/#Waba">Steve's other Waba apps</A>
 * <A HREF="http://www.kagi.com/weyer/#Life">Steve's other Life versions</A>
 * <A HREF="http://faculty.washington.edu/dtlinker/LifeIntro.html">another Waba Life</A>
 * <A HREF="http://www.mindspring.com/~alanh/life/">Other Life patterns and programs</A>
 *
 * ! Newton: class heap: 22000
 * ?? eliminate flicker?  doublebuffer?
**/

public class LifeSW extends MainWindow {
  private static final boolean DEBUG = false; // compiler does strip dead code
  private static final int COMPHT  = 15;
  private static final int COMPWID = 38; // 4 obj w/ 160
  private static final int COMPSEP = 2;

  private static final int ON = 1;
  private static final int OFF = 0;
  private static final int TOGGLE = -1;

  private static final int FIRSTROWCOL = 1000;
  private static final int LASTROWCOL = -1;

  private static final int cellSize = 4; // ?? parameterize
  private static final int sizeX = 32;   // 32 max (i.e., INT)
  private static final int sizeY = 30;
  private static final int DELAY = 20;

  private static final int UNIVTOP  = 20; // 20/120/20 = 160
  private static final int UNIVLEFT = 16; // 16/128/16 = 160
  private static final int UNIVWID  = cellSize*sizeX;
  private static final int UNIVHT   = cellSize*sizeY;

  private int firstRow = FIRSTROWCOL; // to avoid dead rows above
  private int lastRow  = LASTROWCOL;  // to avoid dead rows below
  private int firstCol = FIRSTROWCOL; // ditto for columns
  private int lastCol  = LASTROWCOL;
  private int nGen = 0;  // number of generations (clear =0, nextGeneration +1)
  private int lastX, lastY; // for pen
  private int[] universe;

  private static final int[][] PATTERNS = { // wid,ht,row(s)
    {},                     // dot (same as 1,1,0x1)
    {7,3, 0x20,0x08,0x67},  // acorn
    {4,3, 0x6,0x9,0x6},     // beehive
    {3,1, 0x7},             // blinker
    {2,2, 0x3,0x3},         // block
    {4,3, 0x4,0xE,0xB},     // b-heptomino
    {3,3, 0x2,0x4,0x7},     // glider
    {3,3, 0x7,0x5,0x5},     // pi-heptomino
    {7,3, 0x47,0x72,0x20},  // rabbits
  };
  private static final String[] PATTERN_NAMES = {
    "dot", "acorn", "beehive", "blinker", "block", "b-heptomino",
    "glider", "pi-heptomino", "rabbits",
  };
  private int[] pattern = PATTERNS[0]; // defaults to dot. set by patternPicker

  //private Label genLabel;
  private Button clearButton;
  private Button stepButton;
  private Button randomButton;
  private List patternPicker;
  private Button repeatButton;
  private Timer appTimer = null;
  private Graphics appG = null;
  private Random rand;

  public LifeSW() {
    int objY = 0, objX = 0, objWid = COMPWID;

    // TOP ROW: title, Gen#, Random, pattern
    Title title = new Title("Life");
    title.setRect(objX, objY, this.width, COMPHT+2);
    add(title);

    objX = COMPWID-10;
    //genLabel = new Label("0");
    //genLabel.setRect(objX, objY, objWid, COMPHT);
    //add(genLabel);
    objX += objWid + COMPSEP;

    randomButton = new Button("Random");
    randomButton.setRect(objX, objY, objWid, COMPHT-1);
    add(randomButton);
    objX += objWid + COMPSEP*2;

    patternPicker = new List(PATTERN_NAMES);
    patternPicker.setRect(objX, objY, this.width-objX, COMPHT);
    add(patternPicker);

    // BOTTOM ROW: credit, Clear, Step, Repeat
    objX = 0;
    objY = this.height - COMPHT;

    objWid = COMPWID+1+COMPSEP+24+6; // borrow from left,sep,Clear,Step, Step,Repeat
    Label lbl = new Label("weyer@kagi.com");
    lbl.setRect(objX, objY, objWid, COMPHT);
    add(lbl);
    objX += objWid;

    objWid = COMPWID-12;
    clearButton = new Button("Clear");
    clearButton.setRect(objX, objY, objWid, COMPHT);
    add(clearButton);
    objX += objWid + COMPSEP;

    objWid = objWid-3;
    stepButton = new Button("Step");
    stepButton.setRect(objX, objY, objWid, COMPHT);
    add(stepButton);
    objX += objWid + COMPSEP;

    objWid = COMPWID-3;
    repeatButton = new Button("Repeat");
    repeatButton.setRect(objX, objY, objWid, COMPHT);
    add(repeatButton);

    universe = makeUniverse();
    rand = new Random();
  }

  private void clearAll() { // should this stop??
    universe = makeUniverse();
    nGen = 0;

    // clear former active area
    //update(appG, firstRow,firstCol, lastRow,lastCol);
    firstRow = firstCol = FIRSTROWCOL; // all empty
    lastRow  = lastCol  = LASTROWCOL;
    repaint();
  }

  private void fillRandom() {
    for (int y=0; y < sizeY; y++)
      universe[y] = rand.next(31); //rand.getRandom();
    firstRow = firstCol = 0;
    lastRow = sizeY-1; lastCol = sizeX-1;
    repaint();
  }

  private int[] makeUniverse() {
    int[] univ = new int[sizeY];
    for (int y = 0; y < sizeY; y++)
      univ[y] = 0;
    return univ;
  }

  public void onPaint(Graphics g) {
    update(g, firstRow, firstCol, lastRow, lastCol);
    if (appG == null) { // !!!b
        appG = createGraphics();
        appG.setClip(UNIVLEFT, 0, UNIVWID, UNIVTOP+UNIVHT); // allow text at top
    };
  }

  private void update(Graphics g, int frow, int fcol, int lrow, int lcol) {
    if (DEBUG) System.out.println("update: " + frow + ',' + fcol + ',' + lrow + ',' + lcol);
    g.setColor(255,255,255);  // white
    g.fillRect(UNIVLEFT, UNIVTOP, UNIVWID, UNIVHT); // clear draw area
    g.fillRect(COMPWID-10,1,COMPWID,COMPHT-1);      // clear "#" area
    g.setColor(0,0,0);        // black
    g.drawText(Convert.toString(nGen), COMPWID-10, 2);
    //genLabel.setText(":" + nGen);

    int row;
    for (int y = frow; y <= lrow; y++) {
	row = universe[y];
        if (row != 0) {
          row >>>= fcol;
	  for (int x = fcol; x <= lcol; x++) {
            if ((row & 1) == ON)
              g.fillRect(UNIVLEFT+(x*cellSize), UNIVTOP+(y*cellSize), cellSize, cellSize);
            row >>>= 1;
	  }; // end for-x
        }; // end if
    }; // end for-y
  }

/*
  private int getCell(int row, int x) { // return a bit: 0 or 1. inlined usually
    // ?? left shifting sign bit (x=29) fills 1's (so usually test first)
    // ?? ok w/ unsigned shift
    return (row == 0) ? 0 : (row >>> x) & 1;	// 0 <= x < sizeX. 0 on right
  }
*/

  // set a bit to 0 or 1 (or -1=toggle) & return new row
  private int setCell(int row, int x, int b) {
    int bit1 = 1 << x; 			// 0 <= x < sizeX. 0 on right
    if ((row == 0) && (b != 0))	// 1, -1
      return bit1;
    if ((b != TOGGLE) ? (b == 0) : (row & bit1) != 0)
      return row & ~bit1;	// =0
    return row | bit1;		// =1
  }

  private void nextGeneration() {
    if (firstRow < FIRSTROWCOL) { // only bother if there are some cells
      //universe -- existing
      int[] newuniv = makeUniverse();	// new, empty universe
      int x, y, neighbors, frx, lrx;
      boolean anyUpdate = false;

      int frow = (firstRow > 0) ? firstRow-1 : 0;
      int lrow = (lastRow < sizeY-1) ? lastRow+1 : sizeY-1;
      int fcol = (firstCol > 0) ? firstCol-1 : 0;
      int lcol = (lastCol < sizeX-1) ? lastCol+1 : sizeX-1;

      int newFirstRow = FIRSTROWCOL, newLastRow = LASTROWCOL;
      int newFirstCol = FIRSTROWCOL, newLastCol = LASTROWCOL;

      int n1, n2, n3; // in this layout visually.
      int n4, n5, n6;
      int n7, n8, n9;
      int row1 = 0 , row2 = 0, row3 = universe[frow], newRow = 0;

      // process from top to bottom, from right to left
      for (y = frow; y <= lrow; y++) { // skip some dead rows before/after active area
        row1 = row2; row2 = row3;
        row3 = (y < lrow) ? universe[y+1] : 0;
        if (row1 != 0 || row2 != 0 || row3 !=0) {	// this optimizes for empty row(s)
          frx = FIRSTROWCOL; lrx = LASTROWCOL;
          n1 = (row1 == 0) ? 0 : (row1 >>> fcol) & 1; //getCell(row1,fcol);
          n4 = (row2 == 0) ? 0 : (row2 >>> fcol) & 1; //getCell(row2,fcol);
          n7 = (row3 == 0) ? 0 : (row3 >>> fcol) & 1; //getCell(row3,fcol);
          newRow = n2 = n5 = n8 = n3 = n6 = n9 = 0;
          for (x = fcol; x <= lcol; x++) { // skip some dead columns after/before active area
            if (row1 != 0) {
              n3 = n2; n2 = n1;
              n1 = (x == lcol) ? 0 : (row1 >>> (x+1)) & 1; //getCell(row1,x+1);
              neighbors = n1+n2+n3;
            }
            else
              neighbors = 0;
            if (row2 != 0) {
              n6 = n5; n5 = n4;
              n4 = (x == lcol) ? 0 : (row2 >>> (x+1)) & 1; //getCell(row2,x+1);
              neighbors += n4+n6; // omit center
              };
            if (row3 != 0) {
              n9 = n8; n8 = n7;
              n7 = (x == lcol) ? 0 : (row3 >>> (x+1)) & 1; //getCell(row3,x+1);
              neighbors += n7+n8+n9;
              };
            if (neighbors == 3 || (neighbors == 2 && n5 ==1)) {
              newRow = newRow | (1 << x); //setCell(newRow, x, ON);
              if (x < frx) frx = x; // 1st(min)
              lrx = x;
            };
          }; // end for-x

          anyUpdate = anyUpdate || (row2 != newRow);
          if (lrx > LASTROWCOL) { // something set in row?
            newuniv[y] = newRow;
            newLastRow = y; // last(max)
            if   (y < newFirstRow) newFirstRow =   y; // 1st(min)
            if (frx < newFirstCol) newFirstCol = frx; // min
            if (lrx > newLastCol)  newLastCol  = lrx; // max
          }; // end if
        }; // end if
      }; // end for-y

      if (anyUpdate) {
        universe = newuniv;
        nGen++;
        update(appG, // just active area
          (firstRow < newFirstRow) ? firstRow : newFirstRow,
          (firstCol < newFirstCol) ? firstCol : newFirstCol,
          (lastRow  > newLastRow)  ? lastRow  : newLastRow,
          (lastCol  > newLastCol)  ? lastCol  : newLastCol);

        firstRow = newFirstRow; lastRow = newLastRow;
        firstCol = newFirstCol; lastCol = newLastCol;
      }
      else
        resetTimer(0);  // no change
    }
    else
      resetTimer(0); // empty
  }

  private void resetTimer(int ms) {
    if (ms > 0 && appTimer == null && firstRow < FIRSTROWCOL) { // turn on (only if off & non-empty)
      appTimer = addTimer(ms);
      repeatButton.setText("Stop");
    }
    else if (ms == 0 && appTimer != null) {  // turn off (if on)
      removeTimer(appTimer);
      appTimer = null;
      repeatButton.setText("Repeat");
    };
  }

  public void onEvent(Event ev) {
    Object obj = ev.target;
    int evType = ev.type;
    if (evType == ControlEvent.TIMER)
        nextGeneration();

    else if (obj == this && ev instanceof PenEvent) {
      int x = (((PenEvent) ev).x - UNIVLEFT) / cellSize;
      int y = (((PenEvent) ev).y - UNIVTOP)  / cellSize;
      if (x < 0)
        x = 0;
      else if (x >= sizeX)
        x = sizeX-1;
      if (y < 0)
        y = 0;
      else if (y >= sizeY)
        y = sizeY-1;
if (DEBUG) System.out.println("x: " + x + "; y: " + y);
      if (evType == PenEvent.PEN_DOWN) {
        lastX = x;
        lastY = y;
      }
      else if (evType == PenEvent.PEN_UP)
        setXY(lastX, lastY, x,y);
    }

    else if (evType == ControlEvent.PRESSED) { // Button
      if (obj == stepButton)
        nextGeneration();
      else if (obj == clearButton)
        clearAll();
      else if (obj == repeatButton)
        resetTimer((appTimer == null) ? DELAY : 0); // toggle
      else if (obj == randomButton)
        fillRandom();
      else if (obj == patternPicker)
        pattern = PATTERNS[patternPicker.getSelectedIndex()];
    };
  }

  private void setXY(int x1, int y1, int x2, int y2) {
    int plen= pattern.length;
    boolean rect = (x1 != x2 || y1 != y2);
    if (rect) { // fill rect
      if (x1 > x2) {  // make x1,y1 be upper left; x2,y2 lower right
        x1 = x2;
        x2 = lastX;
      };
      if (y1 > y2) {
        y1 = y2;
        y2 = lastY;
      };
    };
    if (!rect && (plen <= 2 || pattern[1] != plen-2)) // dot
      universe[y1] = setCell(universe[y1], x1, TOGGLE);	// toggle
    else {
      if (!rect) {  // pattern
        y2 = y1+pattern[1]-1; // ylen
        if (y2 >= sizeY) y2 = sizeY-1;  // clip pattern at bottom
        x2 = x1+pattern[0]-1; // xlen
        if (x2 >= sizeX) x2 = sizeX-1;  // clip pattern at right
      };
      int py = 2, row, patrow;

      for (int y = y1; y <= y2; y++) {
        patrow = (rect) ? 0xFFFFFFFF : pattern[py++];
        row = universe[y];
        for (int x = x2; x >= x1; x--) { // pattern reversed
          row = setCell(row, x, patrow & 1);
          patrow >>>= 1;
        }; // end for-x
        universe[y] = row;
      }; // end for-y
    }; // end else

    // expand live area
    if (y1 < firstRow) firstRow = y1;
    if (y2 > lastRow)  lastRow  = y2;
    if (x1 < firstCol) firstCol = x1;
    if (x2 > lastCol)  lastCol  = x2;
    repaint();
  }
}
